# perl script by Kevin Cheung (c) 2001

# subroutines

# print_row(@list)
# prints the row with tabs

# print_row(@list)
sub print_row
{
    for ($j=0; $j < $#_; $j++) {
	print OUTFILE "$_[$j]\t";
        }         
 
    print OUTFILE "$_[$#_]\n";
}

# for sort routine used in top10 selection
my(%count_hash);  # i use a hash to store counts

sub gt_top
{   
  if ($count_hash{$a} > $count_hash{$b}) {
    return 1;
  }
  else {
    return -1;
  }
}

## INPUTFILES
## MULTIFUN, BNUMFORMAT
my(@multifun);
my(%multifun_index); ## hash multifun name to index num
my(%bnumformat);
my(@cluster_x_bnums);
my(@bnums_present); 
my($total_number_of_bnums);

sub load_multifun
{
@multifun = ();
$multifun_index = ();
$index = 1; # ***** 1 indexed array so i can account for $4.S entries.
            #       the reason being that entries not present in the hash
			#       are returned as 0
			#       by one-indexing, i can test to see if functional cat is in the hash

open(DATALIST, $_[0]) ;
while(<DATALIST>) {
    chop($_);    # lop of newline character
    push(@multifun, $_);
    @tmp = split(/\t/, $_);
    $multifun_index{$tmp[0]} = $index;  
    $index = $index + 1;
}
close(DATALIST);
print "loaded multifun.\n";
}

sub load_clusterfile
{
# store classifications for genes in this array
@cluster_x_bnums = ();
$bnums_present = ();

my(@cluster_array) = ();
my($cnum_index);

#get data

open(DATALIST, $_[0]);
while(<DATALIST>) {
    chop($_);    # lop of newline character
    push(@cluster_array, $_);  
}
close(DATALIST);

# iterate through list of clusters
$cnum_index = 0;

# $input_row[0] = e.g. 1
# $input_row[1] = e.g. b2052

@input_row = split(/\t/, $cluster_array[$i]);
$cnum_index = $input_row[0];
$cluster_x_bnums[$cnum_index] = [$input_row[1]];

$bnums_present{$input_row[1]} = 1;
$total_number_of_bnums = 1;

for ($i=1; $i < $#cluster_array + 1; $i++)
{
  @input_row = split(/\t/, $cluster_array[$i]);
  $bnums_present{$input_row[1]} = 1;
  $total_number_of_bnums += 1;

  if ($cnum_index != $input_row[0])
  {
    $cnum_index = $input_row[0];
    $cluster_x_bnums[$cnum_index] = [$input_row[1]];
  }
  else
  {
    $cluster_x_bnums[$cnum_index] = [@{$cluster_x_bnums[$cnum_index]}, $input_row[1]];
  }
}

}

## must load --after-- clusters (because not all bnums may be represented in total set of genes
## that are clustered.

sub load_bnumformat
{
%bnumformat = ();
open(DATALIST, $_[0]);
while(<DATALIST>) {
    chop($_);    # lop of newline character
    @tmp = split(/\t/, $_);
	if ($bnums_present{$tmp[0]} == 1) {
      $bnumformat{$tmp[0]} = [@{$bnumformat{$tmp[0]}}, $tmp[1]];
	}
}
close(DATALIST);
print "loaded bnum_formatted.\n";
}


#### fisher exact test
## given in sokal p.735

########### fisher subroutines ################

# calculating logs a la church
sub logfactorial
{
  my($a);
  my($logtotal);
  my($i); my($j);

  $a = $_[0];
  $logtotal = 0;

  for ($j=1; $j <= $a; $j++) {
     $logtotal = $logtotal + log($j);
  }

  return $logtotal;
};

# calculate fisher-test
# *---*---*
# | a | b |
# *---*---*
# | c | d |
# *---*---*
#
sub calc_fisher_test
{
  my $a, $b, $c, $d;
  my $bt, $ut;
  my $cump;

  $a = $_[0];
  $b = $_[1];
  $c = $_[2];
  $d = $_[3];

  ## want ad - bc to be negative
  if ($a * $d - $b * $c > 0) {
    $t = $a; $a = $b; $b = $t;
    $t = $c; $c = $d; $d = $t; 
  }

  $ut = logfactorial($a + $b) + logfactorial($c + $d) + logfactorial($b + $d) + logfactorial($a + $c) - logfactorial($a + $b + $c + $d); 

  $cump = 0; 
  while (($a >= 0) && ($d >= 0))
  {
     $bt = logfactorial($a) + logfactorial($b) + logfactorial($c) + logfactorial($d);
     $cump = $cump + exp($ut - $bt);

     $a = $a - 1;
     $b = $b + 1;
     $c = $c + 1;
     $d = $d - 1;
  }
  
  return(2 * $cump);
}

## total bnums = 4405 (according to Riley et al)
## but only a subset of that number is actually clustered.

# proper usage:
# calc_fisher_test(observed_matches_in_cluster, observed_matches_outside_cluster, 
#                  # of notmatched in cluster, # of notmatched outside cluster)

######################################
## use hypergeometric
######################################

## calc (n choose x)
sub log_n_choose_x
{
  my $a, $b;
  my $c0, $c1;
  my $i;
 
  $a = $_[0];
  $b = $_[1];

  if ($b == 0) { 0; } ## only 1 way to choose 0 things

  $c0 = 0;
  for ($i=($a - $b + 1); $i <= $a; $i++)
  {
    $c0 = $c0 + log($i);
  }

  $c1 = 0;
  for ($i=1; $i <= $b; $i++)
  {
    $c1 = $c1 + log($i);
  }

  return ($c0 - $c1);
}

## following tavazoie et al (k,f,n,g)
sub calc_hypergeometric
{
  my $f, $g, $n;
  my $cump;
  my $logterm;
  my $i;

  $k = $_[0];
  $f = $_[1];
  $n = $_[2];
  $g = $_[3];

  # the maximum worst case distribution is if
  # all of cluster (n) is your functional category
  # OR if all of functional category is in your cluster.

  $max_allowed = $n;
  if ($n > $f) { $max_allowed = $f; }

  $i = $k;
  $logterm = log_n_choose_x($f, $i) + log_n_choose_x($g - $f, $n - $i) - log_n_choose_x($g, $n);

  $cump = exp($logterm);
  
  for ($i=$k+1; $i <= $max_allowed; $i++)
  {
    $logterm = $logterm + log($f - $i + 1) - log($i) - log($g - $f - ($n - $i + 1) + 1) + log(($n - $i + 1));

    $cump = $cump + exp($logterm);
  }

  return ($cump);
}

###############################################

# MAIN CODE BLOCK

print "\nscore clusters \n";
print "(c) 2001 Kevin Cheung\n";
print "score clusters based on genPROTEC database\n\n";
print "usage: score_clust <clust.txt> max-pval [optional: header bnumformatted-file multifun-file]\n";
print "usage: cluster file must be of the form: [cluster] [bnum]\n";
print "note: requires files bnum_formatted.txt, MultiFun_decode.txt\n";
print "note: make sure last line is a carriage return.\n";
print "note: bnum assignments to functional categories not present\n";
print "      are dumped in the last row as [unspecified] \n";
print "note: max-pval = maximum p-value to report x 100 \n";
print "note: result will be clust.txt.header.fc, clust.txt.header.scores\n";
print "output in clust.txt.scores\n\n";

if (($#ARGV + 1) == 0)
{
    print("Error: need to specify an input file.\n\n");
    exit;
}

if (!(-e $ARGV[0]))
{
    print("Error: inputfile,", $ARGV[0], ", doesn't exist.\n\n");
    exit;
}

$default_mpval = 5;
$default_mpval = $ARGV[1] unless $ARGV[1] <= 0;

## load data files
## important to load clusters first because only then do we 
## know the set of bnums represented
## bnumformat will only then represent those bnums
## multifun counts (in tarray and total) will then reflect the correct counts.

if ($ARGV[2] ne "") {
	$header_tag = "." . $ARGV[2];
}
else
{ 
	$header_tag = "";
}

if ($ARGV[4] ne "") 
{	$multifun_filename = $ARGV[4]; }
else 
{ $multifun_filename = "multifun_decode.txt"; }
if (!(-e $multifun_filename))
    {
      print("Error: inputfile,", $multifun_filename, ", doesn't exist.\n\n");
      exit;
	}
load_multifun($multifun_filename);

load_clusterfile $ARGV[0];

if ($ARGV[3] ne "") 
{ $bnf_filename = $ARGV[3];}
else
{ $bnf_filename = "bnum_formatted.txt"; }
if (!(-e $bnf_filename))
    {
      print("Error: inputfile,", $bnf_filename, ", doesn't exist.\n\n");
      exit;
	}
load_bnumformat($bnf_filename);

## multifun x clusters
my(@tarray); ## sample multifun representation
my(@total); ## total multifun representation
my(@fish_array);
my(@unspecified); ## list of unspecified category matches (e.g. if 
                  ## for a cluster with a bnum, the bnum is a member of 
				  ## a functional category of the 4.S type or isn't in the multifun_idx array
my($errormsg);    ## stores all error info that needs to be in the score file.

$errormsg = "";

####### TABULATION PHASE

# calculate totals
# bnumformat = hash {bnumbers} to {set of functional categories
# multifun_index = hash {functional categ} to 1-indexed (basically a set}
# screens out 4.S entries which for each bnum refers to its -unique- substrate.
# total = {functional categories (0-indexed} to # of bnum entries.

print("calculate totals across all bnums.\n");

for ($i = 0; $i < $#multifun + 1; $i++)
{
  $total[$i] = 0;
}

## what I need to correct for
## Note though that SOME CATEGORIES have no bnum counts!
## look at all bnum keys
while (($key, $value) = each %bnumformat) {
  @bnum_mfs = @$value;
  # for each bnum_mfs, match with index.
  foreach $cat (@bnum_mfs) {
    $idx = $multifun_index{$cat};  ## 1-indexed, so i can check for nonexisted entries   
    if ($idx == 0)
    {
       if ($cat !~ /4\.S.*/) {
         print("couldn't find $cat from $key in functional categories.\n"); 
		 $errormsg = $errormsg . "\nerror: couldn't find $cat from $key in functional categories."; 
         ## 4.S are automatically screened out...
       }
    }
    else
    {
      $total[$idx - 1] += 1;
    }
  }
}

## end calculate totals

print("calculate sample clusters.\n");

$UNREPRESENTED_FC = 1000; ## a really high probability
#initialize tarray
for ($i=0; $i < $#multifun + 1; $i++)
{
  for ($j=0; $j < $#cluster_x_bnums + 1; $j++)
  {
    ${$tarray[$i]}[$j] = 0;
    ${$fish_array[$i]}[$j] = $UNREPRESENTED_FC; ## to flag an error if not all elements are accounted for
  }
}

#initialize array for unspecified
for ($i = 0; $i < $#cluster_x_bnums + 1; $i++)
{
  $unspecified[$i] = "";
}

# for each cluster, tally appropriate entry in tarray
print "tallying cluster ";

for ($i=0; $i < $#cluster_x_bnums + 1; $i++)
{
  @bnums = @{$cluster_x_bnums[$i]};
  print "$i.";
  foreach $bnum (@bnums) {
    # tally the bnum

    @bnum_mfs = @{$bnumformat{$bnum}};
    # for each bnum_mfs, match with index.
    foreach $cat (@bnum_mfs) {
      $idx = $multifun_index{$cat};  ## 1-indexed, so i can check for nonexisted entries   
      if ($idx == 0)
      {
         $unspecified[$i] = $unspecified[$i] . $cat . " ";
      }
      else
      {
        ${$tarray[$idx - 1]}[$i] += 1; ## zero-indexed (functional category counts)
      }
    }
  }
}
print "\n";

print "calculating hypergeometric test.\n";

for ($i=0; $i < $#multifun + 1; $i++)
{     
   if ($total[$i] > 0) {
	   ## only matters if # of genes in category is > 0
	   for ($j = 0; $j < $#cluster_x_bnums + 1; $j++)
	   {
		  ## $#{$cluster_x_bnums[$j]} + 1 = size of cluster
		  ## $total[$i] = representation among entire gene pool
		  ## ${$tarray[$i]}[$j] = representation in cluster
		  ## $total_number_of_bnums = entire gene pool

		  # proper usage:
		  # calc_hypergeometric (k,f,n,g)

		  $a = ${$tarray[$i]}[$j];
		  $b = $total[$i];
		  $c = $#{$cluster_x_bnums[$j]} + 1;
		  $d = $total_number_of_bnums;

		  ${$fish_array[$i]}[$j] = calc_hypergeometric($a,$b,$c,$d);
	   }
   }
}

####### PRINTING STAGE

### print tallied table

$tally_filename = $ARGV[0] . "$header_tag.scores";
open(OUTFILE, ">$tally_filename") || die "could not create $tally_filename\n\n";

print OUTFILE "fc	fc_descript	total	";
for ($i=0; $i < $#cluster_x_bnums + 1; $i++)
{
  print OUTFILE "clust$i	";
}

## fisher test ...
for ($i=0; $i < $#cluster_x_bnums + 1; $i++)
{
  print OUTFILE "fish$i	";
}

print OUTFILE "\n";

for ($i=0; $i < $#multifun + 1; $i++)
{
  # only print if a represented row (total should > be zero)

  if ($total[$i] > 0) 
  { 
  # print clusters
  print OUTFILE "$multifun[$i]	$total[$i]	";
  for ($j = 0; $j < $#cluster_x_bnums + 1; $j++)
  {
    print OUTFILE "${$tarray[$i]}[$j]	";  
  }

  # print fisher test results
  for ($j = 0; $j < $#cluster_x_bnums + 1; $j++)
  {
    $c = ${$fish_array[$i]}[$j];
    print OUTFILE "$c	";
  }
  print OUTFILE "\n";
  #}
  }
}

#print unspecified
print OUTFILE "unspecified			";
for ($j = 0; $j < $#cluster_x_bnums + 1; $j++)
{
  print OUTFILE "$unspecified[$j]	";  
}
print OUTFILE "\n";

close(OUTFILE);

##### PRINT TOPX FILE
### I copy data to new lists because data is easier to play with
### if i want to sort.

$mpval_filename = $ARGV[0] . "$header_tag.top.$default_mpval";
open(OUTFILE, ">$mpval_filename") || die "could not create $mpval_filename\n\n";

print "sorting and getting top $default_mpval for cluster ";

print(OUTFILE "statistics:\n");
$temp = $#cluster_x_bnums + 1;
print(OUTFILE "# of clusters: $temp \n");
print(OUTFILE "# of genes among clusters: $total_number_of_bnums\n");
print(OUTFILE "$errormsg\n\n");

for ($c = 0; $c < $#cluster_x_bnums + 1; $c++)
{
  print "$c.";

  print OUTFILE "cluster $c\n";
  print OUTFILE "------------------------- \n";

  $csize =  $#{$cluster_x_bnums[$c]} + 1;

  print OUTFILE "total cluster size: $csize \n";
  print OUTFILE "------------------------- \n";
  print OUTFILE "cl	rank	fisher-test	function-designation	description	number-in-cluster	number-over-all-clusters	size-of-cluster\n";

  @tmp = ();
  %count_hash = ();

  for ($i=0; $i < $#multifun + 1; $i++)
  {
    $x = $multifun[$i] . "\t" . ${$tarray[$i]}[$c] . "\t" . $total[$i];
    $tmp[$i] = $x;  ## the index

    $count_hash{$x} = ${$fish_array[$i]}[$c]; ## the value; 
											  ## this ensures that only overrepresented stuff gets to the top
										
  }

  @tmp = sort gt_top @tmp;

  for ($i=0; $i < $#multifun + 1; $i++)
  {
	if ($count_hash{$tmp[$i]} < ($default_mpval / 100)) {
	    print OUTFILE "$c	$i	$count_hash{$tmp[$i]}	$tmp[$i]	$csize\n";
	}
  }

  print OUTFILE "-------------------------\n\n\n";
}

close(OUTFILE);

print "\ndone.\n";
exit;

