# perl script by Kevin Cheung (c) 2001

# subroutines

# print_row(@list)
# prints the row with tabs

# print_row(@list)
sub print_row
{
    for ($j=0; $j < $#_; $j++) {
	print OUTFILE "$_[$j]\t";
        }         
 
    print OUTFILE "$_[$#_]\n";
}

# for sort routine used in top10 selection
my(%count_hash);  # i use a hash to store counts

sub gt_top
{   
  if ($count_hash{$a} > $count_hash{$b}) {
    return 1;
  }
  else {
    return -1;
  }
}

## INPUTFILES
## MULTIFUN, BNUMFORMAT
my(@multifun);
my(%multifun_index); ## hash multifun name to index num
my(%bnumformat);
my(@cluster_x_bnums);
my($total_number_of_bnums);

sub load_multifun
{
@multifun = ();
$multifun_index = ();
$index = 1; # ***** 1 indexed array so i can account for $4.S entries.
            #       the reason being that entries not present in the hash
			#       are returned as 0
			#       by one-indexing, i can test to see if functional cat is in the hash

open(DATALIST, $_[0]) ;
while(<DATALIST>) {
    chop($_);    # lop of newline character
    push(@multifun, $_);
    @tmp = split(/\t/, $_);
    $multifun_index{$tmp[0]} = $index;  
    $index = $index + 1;
}
close(DATALIST);
print "loaded multifun.\n";
}

sub load_bnumformat
{
%bnumformat = ();
open(DATALIST, $_[0]);
while(<DATALIST>) {
    chop($_);    # lop of newline character
    @tmp = split(/\t/, $_);
    $bnumformat{$tmp[0]} = [@{$bnumformat{$tmp[0]}}, $tmp[1]];
}
close(DATALIST);
print "loaded bnum_formatted.\n";
}


#### fisher exact test
## given in sokal p.735

########### fisher subroutines ################

# calculating logs a la church
sub logfactorial
{
  my($a);
  my($logtotal);
  my($i); my($j);

  $a = $_[0];
  $logtotal = 0;

  for ($j=1; $j <= $a; $j++) {
     $logtotal = $logtotal + log($j);
  }

  return $logtotal;
};

# calculate fisher-test
# *---*---*
# | a | b |
# *---*---*
# | c | d |
# *---*---*
#
sub calc_fisher_test
{
  my $a, $b, $c, $d;
  my $bt, $ut;
  my $cump;

  $a = $_[0];
  $b = $_[1];
  $c = $_[2];
  $d = $_[3];

  ## want ad - bc to be negative
  if ($a * $d - $b * $c > 0) {
    $t = $a; $a = $b; $b = $t;
    $t = $c; $c = $d; $d = $t; 
  }

  $ut = logfactorial($a + $b) + logfactorial($c + $d) + logfactorial($b + $d) + logfactorial($a + 

$c) - logfactorial($a + $b + $c + $d); 

  $cump = 0; 
  while (($a >= 0) && ($d >= 0))
  {
     $bt = logfactorial($a) + logfactorial($b) + logfactorial($c) + logfactorial($d);
     $cump = $cump + exp($ut - $bt);

     $a = $a - 1;
     $b = $b + 1;
     $c = $c + 1;
     $d = $d - 1;
  }
  
  return(2 * $cump);
}

## total bnums = 4405 (according to Riley et al)
## but only a subset of that number is actually clustered.

# proper usage:
# calc_fisher_test(observed_matches_in_cluster, observed_matches_outside_cluster, 
#                  # of notmatched in cluster, # of notmatched outside cluster)

################# affy database


my(@ABGarray);
my(%gene_to_idx);  # quick index for hash-tables
my(%bnum_to_idx);  # quick index for hash-tables
my(%affy_to_idx);  # quick index for hash-tables

## order
## 0 = affy designation
## 1 = bnum
## 2 = gene name

sub get_affy
{
	my(@row);
	@row = @{$ABGarray[$_[0]]};
	return $row[0];
}

sub get_bnum
{
	my(@row);
	@row = @{$ABGarray[$_[0]]};
	return $row[1];
}

sub get_gene
{
	my(@row);
	@row = @{$ABGarray[$_[0]]};
	return $row[2];
}

sub get_description
{
	my(@row);
	@row = @{$ABGarray[$_[0]]};
	return $row[3];
}

sub load_ABGdb
{
	@ABGarray = ();
	$gene_to_bnum = {};
	$bnum_to_gene = {};
	$affy_to_bnum = {};

    my($index);
    $index = 0;

	open(DATALIST, "affytobnum_idx.txt") ;
	while(<DATALIST>) {
		chop($_);    # lop of newline character
		@tmp = split(/\t/, $_);
		$ABGarray[$index] = [@tmp];
		$affy_to_idx{$tmp[0]} = $index;
		$bnum_to_idx{$tmp[1]} = $index;
		$gene_to_idx{$tmp[2]} = $index;
		$index = $index + 1;
	}
	close(DATALIST);
	print "loaded affy/bnumber/gene database.\n";
}

###############################################

# MAIN CODE BLOCK

print "\nextractfccs \n";
print "(c) 2001 Kevin Cheung\n";
print "extractfccs\n\n";
print "usage: extractfccs<SOM cluster file> <indexfile.txt> [optional: header bnumformatted-file multifun-file]\n";
print "output in clust.txt.keep and clust.txt.extracted\n\n";

if (($#ARGV + 1) == 0)
{
    print("Error: need to specify an input file.\n\n");
    exit;
}

if (!(-e $ARGV[0]))
{
    print("Error: inputfile,", $ARGV[0], ", doesn't exist.\n\n");
    exit;
}

if (!(-e $ARGV[1]))
{
    print("Error: indexfile,", $ARGV[1], ", doesn't exist.\n\n");
    exit;
}

## load data files

if ($ARGV[2] ne "") {
	$header_tag = "." . $ARGV[2];
}
else
{ 
	$header_tag = "";
}

if ($ARGV[4] ne "") 
{	$multifun_filename = $ARGV[4]; }
else 
{ $multifun_filename = "multifun_decode.txt"; }
if (!(-e $multifun_filename))
    {
      print("Error: inputfile,", $multifun_filename, ", doesn't exist.\n\n");
      exit;
	}
load_multifun($multifun_filename);

if ($ARGV[3] ne "") 
{ $bnf_filename = $ARGV[3];}
else
{ $bnf_filename = "bnum_formatted.txt"; }
if (!(-e $bnf_filename))
    {
      print("Error: inputfile,", $bnf_filename, ", doesn't exist.\n\n");
      exit;
	}
load_bnumformat($bnf_filename);

load_ABGdb;

my(%category_filter_hash);

############## load index file

print "getting indexes.\n";
my($ALL_IDENTIFIER);
$ARROW_IDENTIFIER = "->";
$ALL_IDENTIFIER = "(all)";

open(INDEXDATA, $ARGV[1]) || die("$ARGV[1] not found.\n");
while(<INDEXDATA>) {
	chop($_);
	@tmp = split(/\t/, $_);
	$tsz = $tmp[0] . $ARROW_IDENTIFIER . $tmp[1];

	$category_filter_hash{$tsz} = 1;
}

close(INDEXDATA);

############## load cluster data
############## clusters are not sorted, i will also have to sort them
############## will also generate functional category specs.
my(@keep_categories);
my(@extracted_categories);
my($num_rows);
my($num_columns);
my($pos_of_clusters);
my($list_o_cluster_data);
my($total_kept);
$total_kept = 0;

print "parsing clusterfile.\n";

open(CLUSTERDATA, $ARGV[0])|| die("$ARGV[1] not found.\n");
$_ = <CLUSTERDATA>; ## first line is statistics for size of matrix
chop($_);
@tmp = split(/\t/, $_);
$num_rows = $tmp[0];
$num_columns = $tmp[1];
$pos_of_clusters = 2 + $num_columns; ## zero length array-based

$pos = 0;
while(<CLUSTERDATA>) {
	chop($_);
	$list_o_cluster_data[$pos] = $_;
	@tmp = split(/\t/, $_);

	$tmp[0] =~ s/\s*$//;
	$affy_idx = $affy_to_idx{$tmp[0]}; 
	$bnum_idx = get_bnum($affy_idx); # this is the key
	$cluster_identity = $tmp[$pos_of_clusters];

	@tmp_list_o_mfn = @{$bnumformat{$bnum_idx}};
	
	$keepthisone = 1;
	for ($i=0; $i < $#tmp_list_o_mfn + 1 ; $i++) {
		## check if bnum qualifies for ejecting: particular cluster case
		$entry = "$tmp_list_o_mfn[$i]" . $ARROW_IDENTIFIER . "$cluster_identity";
		if ($category_filter_hash{$entry} == 1) {  
			$keepthisone = 0;
			last;
		}
		## check if bnum qualifies for ejecting: A case
		$entry = "$tmp_list_o_mfn[$i]" . $ARROW_IDENTIFIER . $ALL_IDENTIFIER;		
		if ($category_filter_hash{$entry} == 1) {
			$keepthisone = 0;
			last;
		}
	}

	if ($keepthisone == 1) {
		$total_kept = $total_kept + 1;
		$keep_categories[$cluster_identity] = [@{$keep_categories[$cluster_identity]}, $pos]; 
	}
	else
	{
		$extract_categories[$cluster_identity] = [@{$extract_categories[$cluster_identity]}, $pos]; 
	}
	
	$pos = $pos + 1;
}

close(CLUSTERDATA);

##################################################33
print "splitting into two files.\n";

$thefilename = $ARGV[0] . "$header_tag.keep";
open(OUTFILE, ">$thefilename") || die "could not create $thefilename\n";
print OUTFILE "$total_kept	$num_columns\n";

for ($i=0; $i < $#keep_categories + 1; $i++) {
	## spit out clusterdata
	@clustindexes = @{$keep_categories[$i]};
	for ($j=0; $j < $#clustindexes + 1; $j++) {
		$tmp = $list_o_cluster_data[$clustindexes[$j]];
		print OUTFILE "$tmp\n";
	}
}
close(OUTFILE);

$thefilename = $ARGV[0] . "$header_tag.extract";
open(OUTFILE, ">$thefilename") || die "could not create $thefilename\n";
$tmp = $num_rows - $total_kept;
print OUTFILE "total number extracted = $tmp\n\n";

my($fcat_fash); #### fcat hash stores indexes must be 1-indexed
my($pos);

for ($i=0; $i < $#extract_categories + 1; $i++) {
    ## for each cluster:
	##  1) scan to see what functional categories represented
	##  2) store in fcat_hash;

	@clustindexes = @{$extract_categories[$i]};
	if ($extract_categories[$i] eq "") {
		next;
	}

	%fcat_hash = ();
	$pos = 1;

	for ($j=0; $j < $#clustindexes + 1; $j++) {

		$tzs = $list_o_cluster_data[$clustindexes[$j]];
		@tmp = split(/\t/, $tzs);
		$tmp[0] =~ s/\s*$//;
		$affy_idx = $affy_to_idx{$tmp[0]}; 
		$bnum_idx = get_bnum($affy_idx); # this is the key
		$cluster_identity = $tmp[$pos_of_clusters];

		@tmp_list_o_mfn = @{$bnumformat{$bnum_idx}};

		for ($k=0; $k < $#tmp_list_o_mfn + 1 ; $k++) {
			if ($fcat_hash{$tmp_list_o_mfn[$k]} == 0) {
				$fcat_hash{$tmp_list_o_mfn[$k]} = $pos;
				$pos = $pos + 1;
			}
		}
	}

	## now i know which functional categories to represent
	print OUTFILE "in cluster $i\n";
	print OUTFILE "-----------------------------\n\n";

	print OUTFILE "gene\tdescription";
	for ($j=0; $j < $num_columns; $j++) {
		print OUTFILE "\tcond$j";
	}

	## print represented functional categories
	## don't know how hash orders its values so just in case 
	## spit it into a temparray

	@temparray = ();
	while (($key, $value) = each %fcat_hash) {
		$temparray[$value] = $key;
	}
	for ($j=1; $j < $#temparray + 1; $j++) {
		print OUTFILE "\t$temparray[$j]";
	}
	print OUTFILE "\n";

	## now (finally) print out all the clusters with their respective functional categories

	for ($j=0; $j < $#clustindexes + 1; $j++) {

		$tzs = $list_o_cluster_data[$clustindexes[$j]];
		@tmp = split(/\t/, $tzs);
		$tmp[0] =~ s/\s*$//;
		$affy_idx = $affy_to_idx{$tmp[0]}; 
		$bnum_idx = get_bnum($affy_idx); # this is the key
		$cluster_identity = $tmp[$pos_of_clusters];
		$current_cluster = $cluster_identity;

		@temparray = ();
		@tmp_list_o_mfn = @{$bnumformat{$bnum_idx}};

		for ($k=0; $k < $#tmp_list_o_mfn + 1 ; $k++) {
			$temparray[$fcat_hash{$tmp_list_o_mfn[$k]}] = "X";
		}

		for ($k=0; $k < $num_columns + 2; $k++) {
			print OUTFILE "$tmp[$k]\t";
		}

		for ($k=1; $k < $#temparray + 1; $k++) {
			print OUTFILE "$temparray[$k]\t";
		}
		print OUTFILE "\n";
	}
	print OUTFILE "\n\n";

	## now print out the represented functional categories in a list.

	print OUTFILE "category	description\n";
	print OUTFILE "-------------------------------------------\n";

	@temparray = ();
	while (($key, $value) = each %fcat_hash) {
		$tp = $multifun[$multifun_index{$key} - 1]; # 1-indexed
		$temparray[$value] = $key . "\t" . $tp;
	}
	for ($j=1; $j < $#temparray + 1; $j++) {
		print OUTFILE "$temparray[$j]\n";
	}

	print OUTFILE "\n\n";
}

close(OUTFILE);