# process SOM data
# 8/01/2001


# subroutines

# print_row(@list)
# prints the row with tabs

# print_row(@list)
sub print_row
{
    for ($j=0; $j < $#_; $j++) {
	print OUTFILE "$_[$j]\t";
        }         
 
    print OUTFILE "$_[$#_]\n";
}

my(@ABGarray);
my(%gene_to_idx);  # quick index for hash-tables
my(%bnum_to_idx);  # quick index for hash-tables
my(%affy_to_idx);  # quick index for hash-tables

## order
## 0 = affy designation
## 1 = bnum
## 2 = gene name

sub get_affy
{
	my(@row);
	@row = @{$ABGarray[$_[0]]};
	return $row[0];
}

sub get_bnum
{
	my(@row);
	@row = @{$ABGarray[$_[0]]};
	return $row[1];
}

sub get_gene
{
	my(@row);
	@row = @{$ABGarray[$_[0]]};
	return $row[2];
}


sub load_ABGdb
{
	@ABGarray = ();
	$gene_to_bnum = {};
	$bnum_to_gene = {};
	$affy_to_bnum = {};

    my($index);
    $index = 0;

	open(DATALIST, "affytobnum_idx.txt") ;
	while(<DATALIST>) {
		chop($_);    # lop of newline character
		@tmp = split(/\t/, $_);
		$ABGarray[$index] = [@tmp];
		$affy_to_idx{$tmp[0]} = $index;
		$bnum_to_idx{$tmp[1]} = $index;
		$gene_to_idx{$tmp[2]} = $index;
		$index = $index + 1;
	}
	close(DATALIST);
	print "loaded affy/bnumber/gene database.\n";
}

# MAIN CODE BLOCK

print "\process SOM data\n";
print "(c) 2001 Kevin Cheung\n";
print "converts SOM data file to a .dr1, .dr2 file (for matlab) and a .fc (for gc_score).\n\n";
print "usage: process_SOM_data <som_data.txt>\n";
print "output in som_data.txt.dr1, som_data.txt.dr2, and som_data.txt.fc\n\n";


if (($#ARGV + 1) == 0)
{
    print("Error: need to specify an input file.\n\n");
    exit;
}

if (!(-e $ARGV[0]))
{
    print("Error: inputfile,", $ARGV[0], ", doesn't exist.\n\n");
    exit;
}

load_ABGdb;

my($num_columns);
my($num_rows);
my(%hash_o_clusters);
my(@list_o_data);
open(DATALIST, $ARGV[0]) ;
$_ = <DATALIST>;  # should be the header: #rows	#columns
@tmp = split(/\t/, $_);
$num_columns = $tmp[0];
$num_rows = $tmp[1];

while(<DATALIST>) {
	chop($_);    # lop of newline character
	@tmp = split(/\t/, $_);    # there may be trailing whitespace.
	$tmp[0] =~ s/\s*$//;
	$idx = $affy_to_idx{$tmp[0]};
	$hash_o_clusters{$tmp[2 + $num_rows]} = [@{$hash_o_clusters{$tmp[2 + $num_rows]}}, $idx];
	for ($i = 0; $i < $num_rows; $i++) {
	  $list_o_data[$idx]->[$i] = $tmp[2 + $i];
	}
}
close(DATALIST);

open(FCLIST, ">$ARGV[0].fc");
open(DRALIST, ">$ARGV[0].dr1");
open(DR2LIST, ">$ARGV[0].dr2");

while (($key, $value) = each %hash_o_clusters) {
	@x = @{$value};
	for ($i = 0; $i < $#x + 1; $i++) {
		$bnum = get_bnum($x[$i]);
		print FCLIST "$key\t$bnum\n";
	}
}

# dr1 is the cluster
## dr2 is the index

print DRALIST "bnum\tcluster";
for ($i = 0; $i < $num_rows; $i++) {
	print DRALIST "\tcond$i(m)";
}

print DRALIST "\n";

print DR2LIST "cluster\tlow\t\high\n";
$pos = 1;  ## using MATLAB (which is 1-indexed)

# calculate means and standard deviations
while (($key, $value) = each %hash_o_clusters) {
	@x = @{$value};
	## key is the bnum, $value are the $idxs
	## here we map the $idx back to list_o_data

	print DR2LIST "$key\t$pos\t";

	for ($i = 0; $i < $#x + 1; $i++) {


		@y = @{$list_o_data[$x[$i]]};
		$bnum = get_bnum($x[$i]);
		print DRALIST "$bnum\t$key";
		for ($j = 0; $j < $num_rows; $j++) {
			print DRALIST "\t$y[$j]";
		}


		$pos = $pos + 1;

		print DRALIST "\n";
	}

	$tmp = $pos - 1;
	print DR2LIST "$tmp\n";
	
}

close(FCLIST);
close(DRALIST);
close(DR2LIST);

print "done.\n";
exit;
